<?php

class GatewayPaypalV2 extends PaymentGateway
{
	/** @var \GatewayPaypalV2Auth */
	private $auth;

	/** @var \PayPal\Rest\ApiContext */
	private $apiContext;

	const TRANSACTION_ID_KEY = 'txnId';

	public function init()
	{
		require_once __DIR__ . '/PayPal-PHP-SDK/autoload.php';
		$this->authorize($this->config);
		return parent::init();
	}

	private function authorize(stdClass $config = null)
	{
		if (($input = $this->getFormParam('input'))) {
			$this->auth = \GatewayPaypalV2Auth::parseFromString($input);
		} else {
			$clientId = null;
			$clientSecret = null;
			$demo = null;
			if (isset($config->clientId) && $config->clientId) {
				$clientId = $config->clientId;
			}
			if (isset($config->clientSecret) && $config->clientSecret) {
				$clientSecret = $config->clientSecret;
			}
			if (isset($config->demo) && $config->demo) {
				$demo = $config->demo;
			}
			$this->auth = new \GatewayPaypalV2Auth($clientId, $clientSecret, $demo);
		}
	}

	/** @return \PayPal\Rest\ApiContext */
	private function getApiContext()
	{
		if (!$this->apiContext) {
			$this->apiContext = new \PayPal\Rest\ApiContext(
				new \PayPal\Auth\OAuthTokenCredential(
					$this->auth->clientId, // ClientID
					$this->auth->clientSecret // ClientSecret
				)
			);
			$this->apiContext->setConfig(array(
				'mode' => $this->auth->demo ? 'sandbox' : 'live',
			));
		}
		return $this->apiContext;
	}

	public function requestCreatePayment()
	{
		$this->authorize();
		$resp = array();
		try {
			$resp['url'] = $this->getPaymentUrl($_POST);
		} catch (Exception $ex) {
			$resp['error'] = $ex->getMessage();
		}
		return $resp;
	}

	public function createInstantRedirectUrl($formVars)
	{
		try {
			return $this->getPaymentUrl($formVars);
		} catch (Exception $ex) {
			$this->setLastError($ex->getMessage());
		}
		return null;
	}

	private function getPaymentUrl(array $params)
	{
		if (!isset($params['amount']) || !$params['amount']) throw new ErrorException('Amount is not specified');
		if (!isset($params['currency']) || !$params['currency']) throw new ErrorException('Currency is not specified');
		if (!isset($params['description']) || !trim($params['description'])) throw new ErrorException('Description is not specified');
		$transactionId = (isset($params['transactionId']) && $params['transactionId']) ? $params['transactionId'] : md5(microtime());
		$returnUrl = (isset($params['returnUrl']) && $params['returnUrl'])
			? $params['returnUrl'] : getBaseUrl() . 'store-return/Paypal';
		$returnUrl .= '?txnId=' . $transactionId;
		$cancelUrl = (isset($params['cancelUrl']) && $params['cancelUrl'])
			? $params['cancelUrl'] : getBaseUrl() . 'store-cancel/Paypal';
		$cancelUrl .= '?txnId=' . $transactionId;

		$cartData = StoreData::getCartData();

		$payer = new \PayPal\ApiV2\Payer();
		$payer->setPaymentMethod('paypal');

		$payerInfo = new \PayPal\ApiV2\PayerInfo();
		if ($cartData && $cartData->billingInfo) {
			$payerInfo->setEmail($cartData->billingInfo->email);
		}

		$payer->setPayerInfo($payerInfo);

		$itemAmount = new \PayPal\ApiV2\Amount();
		$itemAmount->setValue($params['amount']);
		$itemAmount->setCurrencyCode($params['currency']);

		$items = new \PayPal\ApiV2\ItemList();
		$items->addItem(
			(new \PayPal\ApiV2\Item())
				->setCurrency($params['currency'])
				->setName($params['description'])
				->setDescription($params['description'])
				->setUnitAmount($itemAmount)
				->setQuantity(1)
		);

		$breakdown = new \PayPal\ApiV2\Breakdown();
		$breakdown->setItemTotal($itemAmount);

		$amount = new \PayPal\ApiV2\Amount();
		$amount->setValue($params['amount']);
		$amount->setCurrencyCode($params['currency']);
		$amount->setBreakdown($breakdown);
		$items->setAmount($amount);

		$experienceContext = new \PayPal\ApiV2\ExperienceContext();
		$experienceContext->setCancelUrl($cancelUrl);
		$experienceContext->setReturnUrl($returnUrl);

		$paypal = new \PayPal\ApiV2\Paypal();
		$paypal->setExperienceContext($experienceContext);

		$paymentSource = new \PayPal\ApiV2\PaymentSource();
		$paymentSource->setPaypal($paypal);

		$payment = new \PayPal\ApiV2\Payment();
		$payment->setIntent('CAPTURE')
			->setPayer($payer)
			->setPurchaseUnits(array($items))
			->setPaymentSource($paymentSource);

		$payment->create($this->getApiContext());
		if ($payment->getId()) {
			$_SESSION[self::TRANSACTION_ID_KEY] = $payment->getId();
			if ($order = StoreModuleOrder::findByTransactionId($transactionId)) {
				$order->setExtTransactionId($payment->getId());
				$order->save();
			}
		}
		return $payment->getApprovalLink();
	}

	public function getTransactionId()
	{
		$transactionId = $this->getQueryParam('txnId');
		/* @var $order StoreModuleOrder */
		$order = $transactionId ? StoreModuleOrder::findByTransactionId($transactionId) : null;
		if ($order) {
			return $order->getTransactionId();
		}
		return null;
	}

	public function callback(StoreModuleOrder $order = null)
	{
		if ($order && $order->getExtTransactionId()) {
			$payment = \PayPal\ApiV2\Payment::get($order->getExtTransactionId(), $this->getApiContext());

		} elseif (isset($_SESSION[self::TRANSACTION_ID_KEY])) {
			$payment = \PayPal\ApiV2\Payment::get($_SESSION[self::TRANSACTION_ID_KEY], $this->getApiContext());
			unset($_SESSION[self::TRANSACTION_ID_KEY]);
			$_SESSION[self::TRANSACTION_ID_KEY] = null;
		}

		if (isset($payment) && $payment) {
			if ($payment->getStatus() === 'APPROVED') {
				$payment = $payment->capture($this->getApiContext());
			}
			if ($payment->getStatus() === 'COMPLETED') {
				return true;
			}
		}

		if ($order) {
			$order->setState(StoreModuleOrder::STATE_FAILED);
			$order->save();
		}
		return false;
	}

	public function completeCheckout()
	{
		$transactionId = $this->getQueryParam('txnId');

		/* @var $order StoreModuleOrder */
		$order = $transactionId ? StoreModuleOrder::findByTransactionId($transactionId) : null;
		if ($order && $order->getExtTransactionId()) {
			$payment = \PayPal\ApiV2\Payment::get($order->getExtTransactionId(), $this->getApiContext());

		} elseif (isset($_SESSION[self::TRANSACTION_ID_KEY])) {
			$payment = \PayPal\ApiV2\Payment::get($_SESSION[self::TRANSACTION_ID_KEY], $this->getApiContext());
			unset($_SESSION[self::TRANSACTION_ID_KEY]);
			$_SESSION[self::TRANSACTION_ID_KEY] = null;
		}

		if (isset($payment) && $payment) {
			if ($payment->getStatus() === 'APPROVED') {
				$payment = $payment->capture($this->getApiContext());
			}
			if ($payment->getStatus() === 'COMPLETED') {
				if ($order) {
					$order->setState(StoreModuleOrder::STATE_PAID);
					$order->save();
				}
				return;
			}
		}

		$cancelUrl = getBaseUrl() . 'store-cancel/Paypal';
		header('Location: ' . $cancelUrl, true, 302);
		exit();
	}
}

class GatewayPaypalV2Auth
{

	public $clientId;
	public $clientSecret;
	public $demo;

	private static $instance;

	public function __construct($clientId = null, $clientSecret = null, $demo = false)
	{
		$this->clientId = $clientId;
		$this->clientSecret = $clientSecret;
		$this->demo = (bool)$demo;
	}

	public static function getInstance($shopId = null, $secretKey = null, $demo = false)
	{
		if (!self::$instance) {
			self::$instance = new self($shopId, $secretKey, $demo);
		}
		return self::$instance;
	}

	/**
	 * Decodes auth string and returns object.
	 * @param string|null $string Encoded auth string
	 * @return GatewayPaypalV2Auth
	 */
	public static function parseFromString($string = null)
	{
		for ($i = 0; $i < strlen($string) - 1; $i++) {
			if ($i % 4 == 0) {
				$ch = $string[$i];
				if (is_numeric($ch)) {
					$string[$i] = 10 - intval($ch);
				} else {
					$string[$i] = (strtoupper($ch) === $ch) ? strtolower($ch) : strtoupper($ch);
				}
			}
		}
		$arr = explode(':', base64_decode($string));
		if (count($arr) != 3) {
			return null;
		}
		$obj = new self();
		foreach ($arr as $idx => $part) {
			if ($idx == 0) $obj->clientId = $part;
			else if ($idx == 1) $obj->clientSecret = $part;
			else if ($idx == 2) $obj->demo = $part == '1';
		}
		return $obj;
	}

	/**
	 * Encodes self and returns auth string.
	 * @return string
	 */
	public function parseFromObject()
	{
		$string = base64_encode(implode(':', array($this->clientId, $this->clientSecret, ($this->demo ? '1' : '0'))));
		for ($i = 0; $i < strlen($string) - 1; $i++) {
			if ($i % 4 == 0) {
				$ch = $string[$i];
				if (is_numeric($ch)) {
					$string[$i] = 10 - intval($ch);
				} else {
					$string[$i] = (strtoupper($ch) === $ch) ? strtolower($ch) : strtoupper($ch);
				}
			}
		}
		return $string;
	}

}
